@@ -1,9 +1,8 @@
-#! /usr/bin/env python
 # -*- coding: utf-8 -*-
 #
 # vulture - Find dead code.
 #
-# Copyright (c) 2012-2018 Jendrik Seipp (jendrikseipp@gmail.com)
+# Copyright (c) 2012-2019 Jendrik Seipp (jendrikseipp@gmail.com)
 #
 # Permission is hereby granted, free of charge, to any person obtaining a
 # copy of this software and associated documentation files (the
@@ -38,18 +37,14 @@ import sys
 from vulture import lines
 from vulture import utils
 
-__version__ = '0.29'
+__version__ = '1.2'
 
 DEFAULT_CONFIDENCE = 60
 
-# The ast module in Python 2 trips over "coding" cookies, so strip them.
-ENCODING_REGEX = re.compile(
-    r"^[ \t\v]*#.*?coding[:=][ \t]*([-_.a-zA-Z0-9]+).*?$", flags=re.M)
-
-IGNORED_VARIABLE_NAMES = set(['object', 'self'])
+IGNORED_VARIABLE_NAMES = {'object', 'self'}
 # True and False are NameConstants since Python 3.4.
 if sys.version_info < (3, 4):
-    IGNORED_VARIABLE_NAMES |= set(['True', 'False'])
+    IGNORED_VARIABLE_NAMES |= {'True', 'False'}
 
 
 def _get_unused_items(defined_items, used_names):
@@ -63,6 +58,10 @@ def _is_special_name(name):
     return name.startswith('__') and name.endswith('__')
 
 
+def _match(name, patterns):
+    return any(fnmatchcase(name, pattern) for pattern in patterns)
+
+
 def _is_test_file(filename):
     return any(
         fnmatch(os.path.basename(filename), pattern)
@@ -99,6 +98,8 @@ class Item(object):
     """
     Hold the name, type and location of defined code.
     """
+    __slots__ = ('name', 'typ', 'filename', 'first_lineno',
+                 'last_lineno', 'message', 'confidence')
 
     def __init__(self, name, typ, filename, first_lineno, last_lineno,
                  message='',
@@ -126,6 +127,16 @@ class Item(object):
             utils.format_path(self.filename), self.first_lineno,
             self.message, self.confidence, size_report)
 
+    def get_whitelist_string(self):
+        filename = utils.format_path(self.filename)
+        if self.typ == 'unreachable_code':
+            return ('# {} ({}:{})'.format(
+                self.message, filename, self.first_lineno))
+        else:
+            prefix = '_.' if self.typ in ['attribute', 'property'] else ''
+            return "{}{}  # unused {} ({}:{:d})".format(
+                    prefix, self.name, self.typ, filename, self.first_lineno)
+
     def _tuple(self):
         return (self.filename, self.first_lineno, self.name)
 
@@ -142,7 +153,8 @@ class Item(object):
 class Vulture(ast.NodeVisitor):
     """Find dead code."""
 
-    def __init__(self, verbose=False, ignore_names=None):
+    def __init__(self, verbose=False, ignore_names=None,
+                 ignore_decorators=None):
         self.verbose = verbose
 
         def get_list(typ):
@@ -163,13 +175,14 @@ class Vulture(ast.NodeVisitor):
         self.used_names = get_set('name')
 
         self.ignore_names = ignore_names or []
+        self.ignore_decorators = ignore_decorators or []
 
         self.filename = ''
         self.code = []
         self.found_dead_code_or_error = False
 
     def scan(self, code, filename=''):
-        code = ENCODING_REGEX.sub("", code, count=1)
+        code = utils.sanitize_code(code)
         self.code = code.splitlines()
         self.filename = filename
         try:
@@ -208,7 +221,7 @@ class Vulture(ast.NodeVisitor):
             self._log('Scanning:', module)
             try:
                 module_string = utils.read_file(module)
-            except utils.VultureInputException as err:
+            except utils.VultureInputException as err:  # noqa: F841
                 print(
                     'Error: Could not read file {module} - {err}\n'
                     'Try to change the encoding to UTF-8.'.format(**locals()),
@@ -217,7 +230,7 @@ class Vulture(ast.NodeVisitor):
             else:
                 self.scan(module_string, filename=module)
 
-        unique_imports = set(item.name for item in self.defined_imports)
+        unique_imports = {item.name for item in self.defined_imports}
         for import_name in unique_imports:
             path = os.path.join('whitelists', import_name) + '_whitelist.py'
             if exclude_file(path):
@@ -256,30 +269,18 @@ class Vulture(ast.NodeVisitor):
         return sorted(confidently_unused,
                       key=by_size if sort_by_size else by_name)
 
-    def make_whitelist(self, min_confidence=0, sort_by_size=False):
-        for item in self.get_unused_code(
-                min_confidence=min_confidence, sort_by_size=sort_by_size):
-            if item.typ != 'unreachable_code':
-                prefix = '_.' if item.typ in ['attribute', 'property'] else ''
-                print("{}{}  # unused {} ({}:{:d})".format(
-                        prefix, item.name, item.typ,
-                        utils.format_path(item.filename), item.first_lineno))
-                self.found_dead_code_or_error = True
-        return self.found_dead_code_or_error
-
-    def report(self, min_confidence=0, sort_by_size=False):
+    def report(self, min_confidence=0, sort_by_size=False,
+               make_whitelist=False):
         """
         Print ordered list of Item objects to stdout.
         """
         for item in self.get_unused_code(
                 min_confidence=min_confidence, sort_by_size=sort_by_size):
-            print(item.get_report(add_size=sort_by_size))
+            print(item.get_whitelist_string() if make_whitelist
+                  else item.get_report(add_size=sort_by_size))
             self.found_dead_code_or_error = True
         return self.found_dead_code_or_error
 
-    def _ignore_name(self, name):
-        return any(fnmatchcase(name, pattern) for pattern in self.ignore_names)
-
     @property
     def unused_classes(self):
         return _get_unused_items(
@@ -341,7 +342,7 @@ class Vulture(ast.NodeVisitor):
                 message="unsatisfiable '{name}' condition".format(**locals()),
                 confidence=100)
         else:
-            else_body = getattr(node, 'orelse')
+            else_body = node.orelse
             if utils.condition_is_always_true(node.test) and else_body:
                 self._define(
                     self.unreachable_code, 'else', else_body[0],
@@ -349,11 +350,45 @@ class Vulture(ast.NodeVisitor):
                     message="unreachable 'else' block",
                     confidence=100)
 
+    def _handle_string(self, s):
+        """
+        Parse variable names in format strings:
+
+        '%(my_var)s' % locals()
+        '{my_var}'.format(**locals())
+
+        """
+        # Old format strings.
+        self.used_names |= set(re.findall(r'\%\((\w+)\)', s))
+
+        def is_identifier(name):
+            return bool(re.match(r'[a-zA-Z_][a-zA-Z0-9_]*', name))
+
+        # New format strings.
+        parser = string.Formatter()
+        try:
+            names = [name for _, name, _, _ in parser.parse(s) if name]
+        except ValueError:
+            # Invalid format string.
+            names = []
+
+        for field_name in names:
+            # Remove brackets and contents: "a[0][b].c[d].e" -> "a.c.e".
+            # "a.b.c" -> name = "a", attributes = ["b", "c"]
+            name_and_attrs = re.sub(r'\[\w*\]', '', field_name).split('.')
+            name = name_and_attrs[0]
+            if is_identifier(name):
+                self.used_names.add(name)
+            for attr in name_and_attrs[1:]:
+                if is_identifier(attr):
+                    self.used_attrs.add(attr)
+
     def _define(self, collection, name, first_node, last_node=None,
                 message='', confidence=DEFAULT_CONFIDENCE, ignore=None):
         last_node = last_node or first_node
         typ = collection.typ
-        if (ignore and ignore(self.filename, name)) or self._ignore_name(name):
+        if (ignore and ignore(self.filename, name)) or _match(
+                name, self.ignore_names):
             self._log('Ignoring {typ} "{name}"'.format(**locals()))
         else:
             first_lineno = first_node.lineno
@@ -384,14 +419,26 @@ class Vulture(ast.NodeVisitor):
             self.used_attrs.add(node.attr)
 
     def visit_ClassDef(self, node):
-        self._define(
-            self.defined_classes, node.name, node, ignore=_ignore_class)
+        for decorator in node.decorator_list:
+            if _match(utils.get_decorator_name(decorator),
+                      self.ignore_decorators):
+                self._log('Ignoring class "{}" (decorator whitelisted)'.format(
+                    node.name))
+                break
+        else:
+            self._define(
+                self.defined_classes, node.name, node, ignore=_ignore_class)
 
     def visit_FunctionDef(self, node):
-        for decorator in node.decorator_list:
-            if getattr(decorator, 'id', None) == 'property':
-                self._define(self.defined_props, node.name, node)
-                break
+        decorator_names = [utils.get_decorator_name(
+            decorator) for decorator in node.decorator_list]
+        typ = 'property' if '@property' in decorator_names else 'function'
+        if any(_match(name, self.ignore_decorators)
+               for name in decorator_names):
+            self._log('Ignoring {} "{}" (decorator whitelisted)'.format(
+                    typ, node.name))
+        elif typ == 'property':
+            self._define(self.defined_props, node.name, node)
         else:
             # Function is not a property.
             self._define(
@@ -422,38 +469,13 @@ class Vulture(ast.NodeVisitor):
         elif isinstance(node.ctx, (ast.Param, ast.Store)):
             self._define_variable(node.id, node)
 
-    def visit_Str(self, node):
-        """
-        Parse variable names in format strings:
-
-        '%(my_var)s' % locals()
-        '{my_var}'.format(**locals())
-
-        """
-        # Old format strings.
-        self.used_names |= set(re.findall(r'\%\((\w+)\)', node.s))
-
-        def is_identifier(s):
-            return bool(re.match(r'[a-zA-Z_][a-zA-Z0-9_]*', s))
-
-        # New format strings.
-        parser = string.Formatter()
-        try:
-            names = [name for _, name, _, _ in parser.parse(node.s) if name]
-        except ValueError:
-            # Invalid format string.
-            names = []
-
-        for field_name in names:
-            # Remove brackets and contents: "a[0][b].c[d].e" -> "a.c.e".
-            # "a.b.c" -> name = "a", attributes = ["b", "c"]
-            name_and_attrs = re.sub(r'\[\w*\]', '', field_name).split('.')
-            name = name_and_attrs[0]
-            if is_identifier(name):
-                self.used_names.add(name)
-            for attr in name_and_attrs[1:]:
-                if is_identifier(attr):
-                    self.used_attrs.add(attr)
+    if sys.version_info < (3, 8):
+        def visit_Str(self, node):
+            self._handle_string(node.s)
+    else:
+        def visit_Constant(self, node):
+            if isinstance(node.value, str):
+                self._handle_string(node.value)
 
     def visit_While(self, node):
         self._handle_conditional_node(node, 'while')
@@ -518,8 +540,13 @@ def _parse_args():
     parser.add_argument(
         '--exclude', metavar='PATTERNS', type=csv,
         help='Comma-separated list of paths to ignore (e.g.,'
-        ' "*settings.py,docs/*.py"). {glob_help} A PATTERN without globbing'
-        ' characters is treated as *PATTERN*.'.format(**locals()))
+        ' "*settings.py,docs/*.py"). {glob_help} A PATTERN without glob'
+        ' wildcards is treated as *PATTERN*.'.format(**locals()))
+    parser.add_argument(
+        '--ignore-decorators', metavar='PATTERNS', type=csv,
+        help='Comma-separated list of decorators. Functions and classes using'
+        ' these decorators are ignored (e.g., "@app.route,@require_*").'
+        ' {glob_help}'.format(**locals()))
     parser.add_argument(
         '--ignore-names', metavar='PATTERNS', type=csv, default=None,
         help='Comma-separated list of names to ignore (e.g., "visit_*,do_*").'
@@ -542,10 +569,10 @@ def _parse_args():
 
 def main():
     args = _parse_args()
-    vulture = Vulture(verbose=args.verbose, ignore_names=args.ignore_names)
+    vulture = Vulture(verbose=args.verbose, ignore_names=args.ignore_names,
+                      ignore_decorators=args.ignore_decorators)
     vulture.scavenge(args.paths, exclude=args.exclude)
-    report_func = (vulture.make_whitelist if args.make_whitelist
-                   else vulture.report)
-    sys.exit(report_func(
+    sys.exit(vulture.report(
         min_confidence=args.min_confidence,
-        sort_by_size=args.sort_by_size))
+        sort_by_size=args.sort_by_size,
+        make_whitelist=args.make_whitelist))
